import tensorflow as tf
import tensorflow.keras.backend as K

class AdaptiveDscLoss(tf.keras.losses.Loss):
    """
    Dice coefficient for short, is an F1-oriented statistic used to gauge the similarity of two sets.
    Given two sets A and B, the vanilla dice coefficient between them is given as follows:
        Dice(A, B)  = 2 * True_Positive / (2 * True_Positive + False_Positive + False_Negative)
                    = 2 * |A and B| / (|A| + |B|)
    Math Function:
        U-NET: https://arxiv.org/abs/1505.04597.pdf
        dice_loss(p, y) = 1 - numerator / denominator
            numerator = 2 * \sum_{1}^{t} p_i * y_i + smooth
            denominator = \sum_{1}^{t} p_i + \sum_{1} ^{t} y_i + smooth
        if square_denominator is True, the denominator is \sum_{1}^{t} (p_i ** 2) + \sum_{1} ^{t} (y_i ** 2) + smooth
        V-NET: https://arxiv.org/abs/1606.04797.pdf
    Args:
        smooth (float, optional): a manual smooth value for numerator and denominator.
        square_denominator (bool, optional): [True, False], specifies whether to square the denominator in the loss function.
        with_logits (bool, optional): [True, False], specifies whether the input tensor is normalized by Sigmoid/Softmax funcs.
        ohem_ratio: max ratio of positive/negative, defautls to 0.0, which means no ohem.
        alpha: dsc alpha
    Shape:
        - input: (*)
        - target: (*)
        - mask: (*) 0,1 mask for the input sequence.
        - Output: Scalar loss
    Examples:
        loss = DiceLoss(with_logits=True, ohem_ratio=0.1)
        input = torch.FloatTensor([2, 1, 2, 2, 1])
        input.requires_grad=True
        target = torch.LongTensor([0, 1, 0, 0, 0])
        output = loss(input, target)
        output.backward()
    """
    def __init__(self, smooth=1, square_denominator=False, with_logits=True,ohem_ratio=0.3, alpha=0.01,index_label_position=True, reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE):
        super(AdaptiveDscLoss, self).__init__()
        self.smooth: float = smooth
        self.square_denominator: bool = square_denominator
        self.with_logits: bool = with_logits
        self.ohem_ratio: float = ohem_ratio
        self.alpha: float = alpha
        self.index_label_position = index_label_position
        self.reduction = reduction

    def call(self, y_true, y_pred):
        logits_size = y_pred.shape[-1]
        y_true = tf.reshape(y_true, [-1])
        mask = tf.cast(tf.not_equal(y_true, 0), y_pred.dtype)
        if logits_size != 1:
            y_pred = tf.reshape(y_pred, [-1, logits_size])
            return self._multiple_class(y_pred, y_true, logits_size, mask)
        else:
            y_pred = tf.reshape(y_pred, [-1])
            return self._binary_class(y_pred, y_true, mask)

    @tf.function
    def _compute_dice_loss(self, flat_input, flat_target):
        flat_input = (flat_input+K.epsilon()) ** self.alpha * flat_input
        interection = K.sum(flat_input * flat_target, axis=-1)
        if not self.square_denominator:
            loss = 1 - tf.math.divide_no_nan((2 * interection + self.smooth), (K.sum(flat_input) + K.sum(flat_target) + self.smooth))
        else:
            loss = 1 - tf.math.divide_no_nan((2 * interection + self.smooth), (
                        K.sum(K.square(flat_input)) + K.sum(K.square(flat_target)) + self.smooth))
        return loss

    @tf.function
    def _multiple_class(self, input, target, logits_size, mask=None):
        flat_input = input
        flat_target = K.cast(K.one_hot(target,
                                       num_classes=logits_size), input.dtype) if self.index_label_position else K.cast(
            target, input.dtype)
        # values should be positive in flat_input
        flat_input = tf.math.softmax(flat_input, axis=-1) if self.with_logits else flat_input
        if mask is not None:
            flat_input = flat_input * K.expand_dims(mask, -1)
            flat_target = flat_target * K.expand_dims(mask, -1)
        else:
            mask = K.ones_like(target, dtype=input.dtype)
        loss = None
        if self.ohem_ratio > 0:
            mask_neg = tf.logical_not(K.cast(mask, 'bool'))
            for label_idx in range(logits_size):
                pos_example = K.cast(K.equal(target, label_idx), 'bool')
                neg_example = K.cast(K.not_equal(target, label_idx), 'bool')
                pos_num = K.sum(tf.cast(pos_example, 'int32'))
                neg_num = K.sum(K.cast(mask, 'int32')) - (
                            pos_num - K.sum(K.cast(tf.logical_and(mask_neg, pos_example), 'int32')))
                keep_num = tf.minimum(K.cast(K.cast(pos_num, input.dtype) * self.ohem_ratio / logits_size, 'int32'),
                                      neg_num)
                if keep_num > 0:
                    neg_scores = tf.boolean_mask(flat_input, neg_example)
                    neg_scores_idx = neg_scores[:, label_idx]
                    neg_scores_sort = tf.sort(neg_scores_idx)
                    threshold = neg_scores_sort[-keep_num + 1]
                    cond = ((tf.argmax(flat_input, axis=1) == label_idx) & (flat_input[:, label_idx] >= threshold)) | pos_example
                    ohem_mask_idx = tf.cast(cond, input.dtype)
                    flat_input_idx = flat_input[:, label_idx]
                    flat_target_idx = flat_target[:, label_idx]
                    flat_input_idx = flat_input_idx * ohem_mask_idx
                    flat_target_idx = flat_target_idx * ohem_mask_idx
                else:
                    flat_input_idx = flat_input[:, label_idx]
                    flat_target_idx = flat_target[:, label_idx]
                loss_idx = self._compute_dice_loss(flat_input_idx, flat_target_idx)
                if loss is None:
                    loss = loss_idx
                else:
                    loss += loss_idx
            return loss

        else:
            for label_idx in range(logits_size):
                # pos_example = tf.equal(target, label_idx)
                flat_input_idx = flat_input[:, label_idx]
                flat_target_idx = flat_target[:, label_idx]
                loss_idx = self._compute_dice_loss(flat_input_idx, flat_target_idx)
                if loss is None:
                    loss = loss_idx
                else:
                    loss += loss_idx
            return loss

    @tf.function
    def _binary_class(self, input, target, mask=None):
        flat_target = tf.cast(input, input.dtype)
        flat_input = tf.sigmoid(input) if self.with_logits else input

        if mask is not None:
            mask = tf.cast(mask, input)
            flat_input = flat_input * mask
            flat_target = flat_target * mask
        else:
            mask = tf.ones_like(target)
        if self.ohem_ratio > 0:
            pos_example = target > 0.5
            neg_example = target <= 0.5
            mask_neg_num = mask <= 0.5
            pos_num = K.sum(pos_example) - K.sum(pos_example & mask_neg_num)
            neg_num = K.sum(neg_example)
            keep_num = tf.minimum(tf.cast(pos_num * self.ohem_ratio), neg_num)
            neg_scores = tf.boolean_mask(flat_input, neg_example)
            neg_scores_sort = tf.sort(neg_scores)
            threshold = neg_scores_sort[-keep_num+1]
            cond = (flat_input > threshold) | pos_example
            ohem_mask = tf.cast(cond, 'int32')
            flat_input = flat_input * ohem_mask
            flat_target = flat_target * ohem_mask

        return self._compute_dice_loss(flat_input, flat_target)



import unittest

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Input, Embedding
from tensorflow.keras.models import Model

from tf2crf import CRF, ModelWithCRFLoss, ModelWithCRFLossDSCLoss


def shape_list(x):
    """
    copied from transformers.modeling_tf_utils
    """
    static = x.shape.as_list()
    dynamic = tf.shape(x)
    return [dynamic[i] if s is None else s for i, s in enumerate(static)]


class TFTester(unittest.TestCase):

    # def test_crf_no_kernel(self):
    #     x = np.array([np.random.uniform(0, 1, (3, 9))] * 100)
    #     crf = CRF()
    #     result = crf(x)
    #     assert shape_list(result[0]) == [100, 3]
    #     assert shape_list(result[1]) == [100, 3, 9]
    #     assert shape_list(result[2]) == [100]
    #     assert shape_list(result[3]) == [9, 9]

    def test_crf_kernel(self):
        x = np.array([np.random.uniform(0, 1, (3, 128))] * 100)
        crf = CRF(units=9)
        result = crf(x)
        assert shape_list(result[0]) == [100, 3]
        assert shape_list(result[1]) == [100, 3, 9]
        assert shape_list(result[2]) == [100]
        assert shape_list(result[3]) == [9, 9]

    def test_model_training(self):
        x = np.array([[5, 2, 3] * 3] * 100)
        y = np.array([[1, 2, 3] * 3] * 100)
        model = get_model(9)
        model.fit(x=x, y=y, epochs=5, batch_size=4, validation_split=0.1, verbose=0)

    def test_model_inference(self):
        model = get_model(9)
        model.predict([[5, 2, 3] * 3])

    def test_save_model(self):
        model = get_model(9)
        model.predict([[5, 2, 3] * 3])
        model.save('tests/1')
        model.save_weights('tests/model.h5')

    def test_mixed_precison(self):
        from tensorflow.keras.mixed_precision import experimental as mixed_precision
        x = np.array([[5, 2, 3] * 3] * 100)
        y = np.array([[1, 2, 3] * 3] * 100)
        policy = mixed_precision.Policy('mixed_float16')
        mixed_precision.set_policy(policy)
        model = get_model(9)
        model.fit(x=x, y=y, epochs=5, batch_size=4, validation_split=0.1, verbose=0)

    def test_model_dsc(self):
        x = np.array([[5, 2, 3] * 3] * 100)
        y = np.array([[1, 2, 3] * 3] * 100)
        base_model = get_base_model(9)
        model = ModelWithCRFLossDSCLoss(base_model)
        model.compile(optimizer='adam')
        model.fit(x=x, y=y, epochs=5, batch_size=4, validation_split=0.1, verbose=0)

    def test_mixed_precison_dsc(self):
        from tensorflow.keras.mixed_precision import experimental as mixed_precision
        x = np.array([[5, 2, 3] * 3] * 100)
        y = np.array([[1, 2, 3] * 3] * 100)
        policy = mixed_precision.Policy('mixed_float16')
        mixed_precision.set_policy(policy)
        base_model = get_base_model(9)
        model = ModelWithCRFLossDSCLoss(base_model)
        model.compile(optimizer='adam')
        model.fit(x=x, y=y, epochs=5, batch_size=4, validation_split=0.1, verbose=0)


def get_model(units: int):
    base_model = get_base_model(units)
    model = ModelWithCRFLoss(base_model)
    model.compile(optimizer='adam')
    return model


def get_base_model(units: int):
    inputs = Input(shape=(None,), dtype='int32')
    output = Embedding(10, 20, trainable=True, mask_zero=True)(inputs)
    crf = CRF(units=units, dtype='float32', name='crf')
    output = crf(output)
    base_model = Model(inputs=inputs, outputs=output)
    return base_model


if __name__ == "__main__":
    unittest.main()



from typing import Union

import tensorflow as tf
from tensorflow_addons.text.crf import crf_log_likelihood

from .losses import AdaptiveDscLoss


def unpack_data(data):
    if len(data) == 2:
        return data[0], data[1], None
    elif len(data) == 3:
        return data
    else:
        raise TypeError("Expected data to be a tuple of size 2 or 3.")


class ModelWithCRFLoss(tf.keras.Model):
    """
    Wrapper around the base model for custom training logic.
    Args:
        base_model: The model including the CRF layer
        sparse_target: if the y label is sparse or one-hot, default True
        metric: the metric for training, default 'accuracy'. Warning: Currently tensorflow metrics like AUC need the output and y_true to be one-hot to cauculate, they are not supported.
    """

    def __init__(self, base_model, sparse_target=True, metric: Union[str, object] = 'accuracy'):
        super().__init__()
        self.base_model = base_model
        self.sparse_target = sparse_target
        self.metric = metric
        if isinstance(metric, str):
            if metric == 'accuracy':
                self.metrics_fn = tf.keras.metrics.Accuracy(name='accuracy')
            else:
                raise ValueError('unknown metric name')
        else:
            self.metrics_fn = self.metric
        self.loss_tracker = tf.keras.metrics.Mean(name='loss')

    def call(self, inputs, training=False):
        if training:
            return self.base_model(inputs)
        else:
            return self.base_model(inputs)[0]

    def compute_loss(self, x, y, training=False):
        viterbi_sequence, potentials, sequence_length, chain_kernel = self(x, training=training)
        # we now add the CRF loss:
        crf_loss = -crf_log_likelihood(potentials, y, sequence_length, chain_kernel)[0]
        return viterbi_sequence, sequence_length, tf.reduce_mean(crf_loss)

    def train_step(self, data):
        x, y, sample_weight = unpack_data(data)
        # y : '(batch_size, seq_length)'
        if self.sparse_target:
            assert len(y.shape) == 2
        else:
            y = tf.argmax(y, axis=-1)
        with tf.GradientTape() as tape:
            viterbi_sequence, sequence_length, crf_loss = self.compute_loss(x, y, training=True)
            loss = crf_loss + tf.cast(tf.reduce_sum(self.losses), crf_loss.dtype)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        self.loss_tracker.update_state(loss)
        self.metrics_fn.update_state(y, viterbi_sequence, tf.sequence_mask(sequence_length, y.shape[1]))
        return {"loss": self.loss_tracker.result(), self.metrics_fn.name: self.metrics_fn.result()}

    @property
    def metrics(self):
        return [self.loss_tracker, self.metrics_fn]

    def test_step(self, data):
        x, y, sample_weight = unpack_data(data)
        # y : '(batch_size, seq_length)'
        if self.sparse_target:
            assert len(y.shape) == 2
        else:
            y = tf.argmax(y, axis=-1)
        viterbi_sequence, sequence_length, crf_loss = self.compute_loss(x, y, training=True)
        loss = crf_loss + tf.cast(tf.reduce_sum(self.losses), crf_loss.dtype)
        self.loss_tracker.update_state(loss)
        self.metrics_fn.update_state(y, viterbi_sequence, tf.sequence_mask(sequence_length, y.shape[1]))
        return {"loss_val": self.loss_tracker.result(), f'val_{self.metrics_fn.name}': self.metrics_fn.result()}


class ModelWithCRFLossDSCLoss(tf.keras.Model):
    """
        Wrapper around the base model for custom training logic. And DSC loss to help improve the performance of NER task.
        Args:
            base_model: The model including the CRF layer
            sparse_target: if the y label is sparse or one-hot, default True
            metric: the metric for training, default 'accuracy'. Warning: Currently tensorflow metrics like AUC need the output and y_true to be one-hot to cauculate, they are not supported.
            **kwargs: args for AdaptiveDscLoss.
        """

    def __init__(self, base_model, sparse_target=True, metric: Union[str, object] = 'accuracy', **kwargs):
        super().__init__()
        self.base_model = base_model
        self.sparse_target = sparse_target
        self.metric = metric
        if isinstance(metric, str):
            if metric == 'accuracy':
                self.metrics_fn = tf.keras.metrics.Accuracy(name='accuracy')
            else:
                raise ValueError('unknown metric name')
        else:
            self.metrics_fn = self.metric
        self.loss_tracker = tf.keras.metrics.Mean(name='loss')
        self.dsc = AdaptiveDscLoss(**kwargs)

    def call(self, inputs, training=False):
        if training:
            return self.base_model(inputs)
        else:
            return self.base_model(inputs)[0]

    def compute_loss(self, x, y, sample_weight, training=False):
        viterbi_sequence, potentials, sequence_length, chain_kernel = self(x, training=training)
        # we now add the CRF loss:
        crf_loss = -crf_log_likelihood(potentials, y, sequence_length, chain_kernel)[0]
        ds_loss = self.dsc(y, potentials, y)
        if sample_weight is not None:
            crf_loss = crf_loss * sample_weight[0]
            ds_loss = ds_loss * sample_weight[1]
        return viterbi_sequence, sequence_length, tf.reduce_mean(crf_loss), tf.reduce_mean(ds_loss)

    def train_step(self, data):
        x, y, sample_weight = unpack_data(data)
        # y : '(batch_size, seq_length)'
        if self.sparse_target:
            assert len(y.shape) == 2
        else:
            y = tf.argmax(y, axis=-1)

        with tf.GradientTape() as tape:
            viterbi_sequence, sequence_length, crf_loss, ds_loss = self.compute_loss(x, y, sample_weight, training=True)
            loss = crf_loss + ds_loss + tf.cast(tf.reduce_sum(self.losses), crf_loss.dtype)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        self.loss_tracker.update_state(loss)
        self.metrics_fn.update_state(y, viterbi_sequence, tf.sequence_mask(sequence_length, y.shape[1]))
        return {"loss": self.loss_tracker.result(), self.metrics_fn.name: self.metrics_fn.result()}

    @property
    def metrics(self):
        return [self.loss_tracker, self.metrics_fn]

    def test_step(self, data):
        x, y, sample_weight = unpack_data(data)
        viterbi_sequence, sequence_length, crf_loss, ds_loss = self.compute_loss(x, y, sample_weight, training=True)
        loss = crf_loss + ds_loss + tf.cast(tf.reduce_sum(self.losses), crf_loss.dtype)
        self.loss_tracker.update_state(loss)
        self.metrics_fn.update_state(y, viterbi_sequence, tf.sequence_mask(sequence_length, y.shape[1]))
        return {"loss_val": self.loss_tracker.result(), f'val_{self.metrics_fn.name}': self.metrics_fn.result()}



import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow.keras.backend as K


class CRF(tf.keras.layers.Layer):
    """
    Conditional Random Field layer (tf.keras)
    `CRF` can be used as the last layer in a network (as a classifier). Input shape (features)
    must be equal to the number of classes the CRF can predict (a linear layer is recommended).

    Args:
        units: Positive integer, dimensionality of the output space. If it is None, the last dim of input must be num_classes
        chain_initializer: the initialize method for transitions, default orthogonal.
        boundary_initializer: the initialize method for boundary, default zero
        use_boundary: whether to use boundary, default true
        regularizer: Regularizer for crf transitions, can be 'l1', 'l2' or other tensorflow regularizers.

    Input shape:
        nD tensor with shape `(batch_size, sentence length, features)` or `(batch_size, sentence length, num_classes)`.

    Output shape:
        in training:
            viterbi_sequence: the predicted sequence tags with shape `(batch_size, sentence length)`
            inputs: the input tensor of the CRF layer with shape `(batch_size, sentence length, num_classes)`
            sequence_lengths: true sequence length of inputs with shape `(batch_size)`
            self.transitions: the internal transition parameters of CRF with shape `(num_classes, num_classes)`
        in predicting:
            viterbi_sequence: the predicted sequence tags with shape `(batch_size, sentence length)`

    Masking
        This layer supports keras masking for input data with a variable number
        of timesteps. To introduce masks to your data,
        use an embedding layer with the `mask_zero` parameter
        set to `True` or add a Masking Layer before this Layer
    """

    def __init__(self, units=None, chain_initializer="orthogonal", boundary_initializer='zeros', use_boundary=True, regularizer=None, **kwargs):
        super(CRF, self).__init__(**kwargs)
        self.chain_initializer = tf.keras.initializers.get(chain_initializer)
        self.boundary_initializer = tf.keras.initializers.get(boundary_initializer)
        self.chain_initializer = tf.keras.initializers.get(chain_initializer)
        self.regularizer = tf.keras.regularizers.get(regularizer)
        self.transitions = None
        self.supports_masking = True
        self.mask = None
        self.units = units
        self.use_boundary = use_boundary
        self.accuracy_fn = tf.keras.metrics.Accuracy()

    def get_config(self):
        config = {
            "chain_initializer": tf.keras.initializers.serialize(
                self.chain_initializer),
            "regularizer": self.regularizer,
            "use_boundary": self.use_boundary,
            "boundary_initializer": tf.keras.initializers.serialize(
                self.boundary_initializer
            ),
            "units": self.units,
        }
        base_config = super(CRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        if self.units:
            units = self.units
        else:
            units = input_shape[-1]
        self.transitions = self.add_weight(
            name="transitions",
            shape=[units, units],
            initializer=self.chain_initializer,
            regularizer=self.regularizer
        )

        if self.use_boundary:
            self.left_boundary = self.add_weight(
                shape=(units,),
                name="left_boundary",
                initializer=self.boundary_initializer,
            )
            self.right_boundary = self.add_weight(
                shape=(units,),
                name="right_boundary",
                initializer=self.boundary_initializer,
            )
        if self.units is not None:
            self.dense = tf.keras.layers.Dense(units, dtype=self.dtype)
        else:
            self.dense = lambda x: tf.cast(x, dtype=self.dtype)

    def call(self, inputs, mask=None, training=False):
        potentials = self.dense(inputs)
        if mask is not None:
            if tf.keras.backend.ndim(mask) != 2:
                raise ValueError("Input mask to CRF must have dim 2 if not None")

        if mask is not None:
            left_boundary_mask = self._compute_mask_left_boundary(mask)
            first_mask = left_boundary_mask[:, 0]
            if first_mask is not None and tf.executing_eagerly():
                no_left_padding = tf.math.reduce_all(first_mask)
                left_padding = not no_left_padding
                if left_padding:
                    raise NotImplementedError(
                        "Currently, CRF layer do not support left padding"
                    )
        # appending boundary probability info
        if self.use_boundary:
            potentials = self.add_boundary_energy(
                potentials, mask, self.left_boundary, self.right_boundary
            )
        sequence_length = self._get_sequence_length(inputs, mask)
        viterbi_sequence, _ = tfa.text.crf_decode(
            potentials, self.transitions, sequence_length
        )
        return viterbi_sequence, potentials, sequence_length, self.transitions

    def _get_sequence_length(self, input_, mask):
        """Currently underline CRF fucntion (provided by
        tensorflow_addons.text.crf) do not support bi-direction masking (left
        padding / right padding), it support right padding by tell it the
        sequence length.
        this function is compute the sequence length from input and
        mask.
        """
        if mask is not None:
            sequence_length = self.mask_to_sequence_length(mask)
        else:
            # make a mask tensor from input, then used to generate sequence_length
            input_energy_shape = tf.shape(input_)
            raw_input_shape = tf.slice(input_energy_shape, [0], [2])
            alt_mask = tf.ones(raw_input_shape)

            sequence_length = self.mask_to_sequence_length(alt_mask)

        return sequence_length

    def mask_to_sequence_length(self, mask):
        """compute sequence length from mask."""
        sequence_length = tf.reduce_sum(tf.cast(mask, tf.int64), 1)
        return sequence_length

    @staticmethod
    def _compute_mask_right_boundary(mask):
        """input mask: 0011100, output right_boundary: 0000100."""
        # shift mask to left by 1: 0011100 => 0111000
        offset = 1
        left_shifted_mask = tf.concat(
            [mask[:, offset:], tf.zeros_like(mask[:, :offset])], axis=1
        )

        # NOTE: below code is different from keras_contrib
        # Original code in keras_contrib:
        # end_mask = K.cast(
        #   K.greater(self.shift_left(mask), mask),
        #   K.floatx()
        # )
        # has a bug, confirmed
        # by the original keras_contrib maintainer
        # Luiz Felix (github: lzfelix),

        # 0011100 > 0111000 => 0000100
        right_boundary = tf.math.greater(
            tf.cast(mask, tf.int32), tf.cast(left_shifted_mask, tf.int32)
        )

        return right_boundary

    @staticmethod
    def _compute_mask_left_boundary(mask):
        """input mask: 0011100, output left_boundary: 0010000."""
        # shift mask to right by 1: 0011100 => 0001110
        offset = 1
        right_shifted_mask = tf.concat(
            [tf.zeros_like(mask[:, :offset]), mask[:, :-offset]], axis=1
        )

        # 0011100 > 0001110 => 0010000
        left_boundary = tf.math.greater(
            tf.cast(mask, tf.int32), tf.cast(right_shifted_mask, tf.int32)
        )

        return left_boundary

    def add_boundary_energy(self, potentials, mask, start, end):
        def expand_scalar_to_3d(x):
            # expand tensor from shape (x, ) to (1, 1, x)
            return tf.reshape(x, (1, 1, -1))

        start = tf.cast(expand_scalar_to_3d(start), potentials.dtype)
        end = tf.cast(expand_scalar_to_3d(end), potentials.dtype)
        if mask is None:
            potentials = tf.concat(
                [potentials[:, :1, :] + start, potentials[:, 1:, :]], axis=1
            )
            potentials = tf.concat(
                [potentials[:, :-1, :], potentials[:, -1:, :] + end], axis=1
            )
        else:
            mask = tf.keras.backend.expand_dims(tf.cast(mask, start.dtype), axis=-1)
            start_mask = tf.cast(self._compute_mask_left_boundary(mask), start.dtype)

            end_mask = tf.cast(self._compute_mask_right_boundary(mask), end.dtype)
            potentials = potentials + start_mask * start
            potentials = potentials + end_mask * end
        return potentials

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[:2]
        return output_shape

    def compute_mask(self, input_, mask=None):
        """keep mask shape [batch_size, max_seq_len]"""
        return mask


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ['tensorflow>=2.1.0', 'tensorflow_addons>=0.8.2']

setup_requirements = ["bumpversion"]

setup(
    name='tf2crf',
    version='0.1.32',
    description='a crf layer for tensorflow 2 keras',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT License',
    url='https://github.com/xuxingya/tf2crf',
    author='xingya.xu',
    author_email='xingya.xu@gmail.com',
    install_requires=requirements,
    setup_requires=setup_requirements,
    packages=find_packages()
)
